/**
 *
 * Mozilla Public License Notice
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1  (the "License"); you may  not use this  file except in
 * compliance with the  License. You may obtain a  copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the  License  for  the   specific  language  governing  rights  and
 * limitations under the License.
 *
 * The  Original Code and  Idea is  the Hit-a-Hint  Mozilla extension.
 * The Initial  Developer of the Original  Code and the  Idea is Pekka
 * P. Sillanpaa.  Portions created  by Initial Developer are Copyright
 * (C) 2004.  All Rights Reserved.
 *
 * Contributor(s): Pekka Sillanpaa, Paul Stone
 *
 */

// surfkeys needs this

function hit_a_hint() {

  var hintElemArray = null; // array of hint elements
  var hintableElemArray = null;
  var elemArray = null;
  var linkNumString = ""; // the typed link number is in this string
  var quickDisabled = false;
  var regenElemList = false;
  var isFirstTime = true;
  var areCoordsValid = true;
  var docCount = 0; // counter for frames in the document
  var hahFrameArray = null;
  var hahFAYTState = null;

  var isMagicKeyPressed = false;

  var isHahModeEnabled = false; // is typing mode on


  // not (yet?) in preferences
  var BACKSPACE_KEY = 8; // backspace
  var FINISH_KEY = 13; // enter
  var STOP_KEY = 27; // esc

  var HINT_PREFIX = 'hint'; // prefix for the hint id

  var loads = 0;

  // for BECL (background element coordidate loader)
  var bECLwin = 0;
  var bECLtag = 0;
  var bECLelem = 0;

  var getHintableElemsRunning = false;

  var oWindowDocument = null;

  var toggled = false;
  var tempDisabled = false;

  var BECLid = null;

  var prefs = Components.classes["@mozilla.org/preferences-service;1"].
    getService(Components.interfaces.nsIPrefService);
  prefs = prefs.getBranch("extensions.hah.");

  var magicDoublePress = false;

  ////////////////////////////////////////////////////////////////////////////////
  // logging
  ////////////////////////////////////////////////////////////////////////////////

  var gConsoleService = Components.classes['@mozilla.org/consoleservice;1']
    .getService(Components.interfaces.nsIConsoleService);


  this.QuickToggle = quickToggle;
  this.RegenHints = regenHints;
  this.BlurFocused = blurFocused;

  function logMessage(aMessage) {
    gConsoleService.logStringMessage('hah: ' + aMessage);
  };

  ////////////////////////////////////////////////////////////////////////////////
  // configuration and initialization related functions
  ////////////////////////////////////////////////////////////////////////////////

  function load() {

    if (!prefs.getBoolPref("enabled")) return;

    var currloc = window._content.location.href;
  
    quickDisabled = false;

    var siteArray = "";

    if(!quickDisabled) setKeyEvents(true);
    
    if (isHahModeEnabled) removeHints();

    isHahModeEnabled = false;

    areCoordsValid = false;

    window.setTimeout("hah.startHintableElemsGetter()", 10);
    oWindowDocument = window.document;

    toggled = false;
  };

  function setKeyEvents(enabled) {
    if (enabled) {
      window.addEventListener("keydown", onKeydown, true);
      window.addEventListener("keyup", onKeyup, true);
      window.addEventListener("keypress", onKeypress, true);
    } else {
      window.removeEventListener("keydown", onKeydown, true);
      window.removeEventListener("keyup", onKeyup, true);
      window.removeEventListener("keypress", onKeypress, true);
    }
  }

  function blurFocused() {
    var elt = document.commandDispatcher.focusedElement;
    if (elt != null) elt.blur();
    //  getDocument().body.focus();
  }

  function resize() {
    areCoordsValid = false;
    if (isHahModeEnabled) {
      removeHints();
      showHints();
    }
  }

  function quickToggle() {
    toggled = true;

    if (quickDisabled) {
      quickDisabled = false;
      setKeyEvents(true);
    }
    else {
      quickDisabled = true;
      setKeyEvents(false);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////
  // hint activating and loading related functions
  ////////////////////////////////////////////////////////////////////////////////

  function showHints() {
    generateHints();
  }

  function removeHints() {
    for (var i = 0; i < hintElemArray.length; i++) {
      var hint = hintElemArray[i];
      //hint.ownerDocument.body.removeChild(hint);    
      hintElemArray[i].parentNode.removeChild(hint);
    }
  }


  /**
   * generateHints()
   *
   * iterates the hintableElemArray (generated in getHintableItems)
   * and create a hint element for each visible hintable element
   * 
   * hintableElemArray is a three-dimensional array, where first
   * array is for frames, second for tags, and third for elements
   *
   */

  function generateHints() { 
    var linkCount = 0;

    var doc, styleString, hintElemSpan;

    var area = new Array(4);

    hintElemArray = new Array();
    elemArray = new Array();

    var win = window._content;

    styleString = prefs.getCharPref("hints.style");
    
    var elem, cs, hintText;

    logMessage(hintableElemArray.length)

    for (var j = 0; j < hintableElemArray.length; j++) {
      // move the area  a little bit to up-left so that also links
      // that are just a bit hidden are shown

      area[0] = win.pageXOffset - 5;
      area[1] = win.pageYOffset - 5;
      area[2] = area[0] + win.innerWidth;
      area[3] = area[1] + win.innerHeight;
    
      doc = win.document;
    
      hintElemSpan = doc.createElement('SPAN');
      hintElemSpan.setAttribute('style', styleString);
      hintElemSpan.style.position = 'absolute';

      for (var k = 0; k < hintableElemArray[j].length; k++) {
	for (var i = 0; i < hintableElemArray[j][k].length; i++) {
	  elem = hintableElemArray[j][k][i];

	  if (!areCoordsValid) {
	    //	    getElemCoords(elem);
	  }
	
	  //	  if (elem.absoY > area[3] || elem.absoY < area[1] || elem.absoX > area[2] || elem.absoX < area[0])
	  //	    continue;


	  if (elem.offsetY > area[3] || elem.offsetY < area[1] || elem.offsetX > area[2] || elem.offsetX < area[0])
	    continue;
	
	  if (elem.offsetWidth == 0 && elem.offsetHeight == 0)
	    continue;


	  cs = doc.defaultView.getComputedStyle(elem, '');
	
	  if (cs.visibility == "hidden")
	    continue;
	
	  var hintElem = hintElemSpan.cloneNode(false);
	  hintText = formatHint(linkCount);
	
	  //	if (elem.offsetParent)
	  //	  logMessage(hintText + " " + elem.offsetParent.innerHTML);  

	  hintElem.appendChild(doc.createTextNode(hintText));
	  //hintElem.style.left = elem.absoX + 'px';
	  //hintElem.style.top = elem.absoY + 'px';

	  //doc.body.appendChild(hintElem);
	  //	  elem.offsetParent.appendChild(hintElem);
	  if (elem.parentNode) {
	    elem.parentNode.insertBefore(hintElem, elem);
	  }
	  else {
	    elem.append(hintElem)
	  }

	  elemArray[linkCount] = elem;
	  hintElemArray[linkCount] = hintElem; // save the hintelem so that it can be referenced
	  linkCount++;
	}
      }
 
      win = hahFrameArray[j];
    }
  
    areCoordsValid = true;
    logMessage("out of here");
  }
  
  this.startHintableElemsGetter = function() {
    getHintableElems();
  };

  /**
   * getHintableItems
   *
   * gets all elements listed in the tag list in the preferences
   * this is done for every frame recursively as well
   *
   */
  function getHintableElems() {
    getHintableElemsRunning = true;

    var linkCount = 0;
    var hintableElems = null;


    var elem = null, cs = null, hintElem = null;

    var tags = prefs.getCharPref("hints.tags");
    var tag;  

    if (tags.length == 0) {
      return;
    }

    var docCount = 0;

    var win = window._content;

    hintableElemArray = new Array();
  
    hahFrameArray = new Array();

    var i = 0;

    for (i = 0; i < win.frames.length; i++)
      hahFrameArray.push(win.frames[i]);

    while (win) {
      var doc = win.document;
    
      var hintableTagArray = new Array(); 
    
      var res = doc.evaluate(tags, doc, null, XPathResult.ANY_TYPE, null);

      hintableElems = new Array();

      var e = res.iterateNext();

      while (e) {
	hintableElems.push(e);
	e = res.iterateNext();
      }
    
      hintableTagArray.push(hintableElems);

      hintableElemArray.push(hintableTagArray);
    
      win = hahFrameArray[docCount++];
       
      if (win)
	for (i = 0; i < win.frames.length; i++)
	  hahFrameArray.push(win.frames[i]);
    }

    startBackgroundElemCoordLoader();

    getHintableElemsRunning = false;
  }

  /**
   * Starts the coordinateloader in background, because this reduce
   * the time of hints appearing first time.
   * Getting the elements' absolute coordinates is an expensive job,
   * and in pages with thousands of links this feature adds convenience
   '*/
  function startBackgroundElemCoordLoader() {
    bECLwin = 0;
    bECLtag = 0;
    bECLelem = 0;

    if (BECLid != null) {
      //window.clearTimeout(BECLid);
    }
    //  BECLid = window.setTimeout("backgroundElemCoordLoader()", 1000);
  }

  /**
   * This function calls recursively itself in the background until
   * all the elements' coordinates have been loaded
   *
   */
  function backgroundElemCoordLoader() { // "BECL"
    //  logMessage(bECLwin + ","+bECLtag+","+bECLelem);
  
    if (hintableElemArray[bECLwin].length == 0) {
      bECLtag = 0;
      bECLelem = 0; 
      bECLwin++;
    } else {
      var elem = hintableElemArray[bECLwin][bECLtag][bECLelem];
      getElemCoords(elem);
    
      if (bECLelem < hintableElemArray[bECLwin][bECLtag].length - 1) {
	bECLelem++;
      } else if (bECLtag < hintableElemArray[bECLwin].length - 1) {
	bECLelem = 0;
	bECLtag++;
      } else if (bECLwin < hintableElemArray.length - 1) {
	bECLtag = 0;
	bECLelem = 0;
	bECLwin++;
      } else {
	BECLid = null;
	areCoordsValid = true;
	//      logMessage("Coordinates pre-loaded");
	return;
      }
    }
    if (areCoordsValid || isHahModeEnabled) {
      BECLid = null;    
    } else {
      window.setTimeout("backgroundElemCoordLoader()", 1);  
    }
  }

  function formatHint(hintNum) {
    var numbers = prefs.getCharPref("keys.numbers");
    var str = hintNum.toString(numbers.length); // turn hintNum into a base(length) number
  
    if (prefs.getBoolPref("hints.forceNumeric")) {
      return str;
    } else {
      // map the number onto the chars in the numbers string
      var result = '';
      for (var i=0; i<str.length; i++)
	result += (numbers.charAt(parseInt(str[i], numbers.length))).toUpperCase();;
      return result;
    }
  }

  function getElemCoords(elem) {
    if (elem.absoX == undefined || !areCoordsValid) {
      var elemTemp = elem;
      elem.absoX = 0;
      elem.absoY = 0;
      while (elemTemp.tagName != 'BODY') {
	elem.absoX += elemTemp.offsetLeft;
	elem.absoY += elemTemp.offsetTop;
	elemTemp = elemTemp.offsetParent;

	if (elemTemp == null)
	  break;

	if (elemTemp.absoX != undefined) {
	  elem.absoX += elemTemp.absoX;
	  elem.absoY += elemTemp.absoY;
	  break;
	}
      }
    }

    if (elem.tagName == 'A') {
      var t = elem.firstChild; // we use image coordinates instead of the anchor
      if (t != null && t.tagName != undefined && t.tagName == 'IMG') {
	elem.absoX += t.offsetLeft - elem.offsetLeft;
	elem.absoY += t.offsetTop - elem.offsetTop;
      } 
    } else if (elem.tagName == 'AREA') {
      // find imagemap's coordinates
      var coords = elem.getAttribute("coords").split(",");
      elem.absoX += Number(coords[0]);
      elem.absoY += Number(coords[1]);
    }
  }

  function changeHintFocus(linkNumString, oldLinkNumString) {
    var nums = prefs.getCharPref("keys.numbers");
    var hintId = parseInt(linkNumString, nums.length);
    var oldHintId = parseInt(oldLinkNumString, nums.length);

    var styleString = prefs.getCharPref("hints.style");
    var styleStringFocus = prefs.getCharPref("hints.styleFocus");

    var xTemp, yTemp;

    if (oldLinkNumString.length > 0) {
      hintElem = hintElemArray[oldHintId];
      if (elemArray[oldHintId].tagName == 'FRAME' || elemArray[oldHintId].tagName == 'IFRAME') {
	hintElem.style.backgroundColor = 'ButtonFace';
      } else {
	xTemp = hintElem.style.left;
	yTemp = hintElem.style.top;
	hintElem.setAttribute('style', styleString);
	hintElem.style.position = 'absolute';
	hintElem.style.left = xTemp;
	hintElem.style.top = yTemp;
      }
    }
    if (linkNumString.length > 0) {
      hintElem = hintElemArray[hintId]; // TODO: has no properties??
      if (elemArray[hintId].tagName == 'FRAME' || elemArray[hintId].tagName == 'IFRAME') {
	hintElem.style.backgroundColor = 'ButtonShadow';
      } else {
	xTemp = hintElem.style.left;
	yTemp = hintElem.style.top;
	hintElem.setAttribute('style', styleStringFocus);
	hintElem.style.position = 'absolute';
	hintElem.style.left = xTemp; //(parseInt(xTemp) - (parseInt(widthTemp2) - parseInt(widthTemp)) / 2) + "px";
	hintElem.style.top = yTemp; //(parseInt(yTemp) - (parseInt(heightTemp2) - parseInt(heightTemp)) / 2) + "px";
      }
    }
  }

  function regenHints() {
    areCoordsValid = false;
  }


  ////////////////////////////////////////////////////////////////////////////////
  // basic functionality
  ////////////////////////////////////////////////////////////////////////////////

  /**
   * Enables the HaH-mode by showing the hints and prepare for input the
   * hint numbers
   *
   * @author Pekka Sillanpaa
   * @param event that caused the mode to change
   * @return -1 if already enabled
   */

  function enableHahMode(event) {
    if (isHahModeEnabled) 
      return -1;

    if (oWindowDocument != window._content.document) {
      areCoordsValid = false;
      getHintableElems();
    }

    oWindowDocument = window._content.document;

    linkNumString = '';
    isHahModeEnabled = true;
    showHints();

    //hahFAYTState = gUseTypeAheadFind; // disable FAYT
    //if (gUseTypeAheadFind) gUseTypeAheadFind = false;

    return 0;
  }

  /**
   * Disables the HaH-mode by hiding the hints and disabling the input mode
   *
   * @author Pekka Sillanpaa
   * @param event that caused the mode to change
   * @param action = true if something is to be clicked
   *                 false if cancel
   * @return -1 if already disabled
   */
  function disableHahMode(event, action) {
    if (!isHahModeEnabled) 
      return -1;

    removeHints();

    isHahModeEnabled = false;
  
    var nums = prefs.getCharPref("keys.numbers");
    var doc = window.document;
  
    var hintId = parseInt(linkNumString, nums.length);
    linkNumString = '';

    gUseTypeAheadFind = hahFAYTState; // re-enable FAYT if it was enabled

    if (hintId >= 0 && hintId < elemArray.length && action) {
      var x = 0;
      var y = 0;
      elem = elemArray[hintId];

      //    var x = elem.absoX;
      //    var y = elem.absoY;

      var elemTagName = elem.tagName;

      if (elemTagName == 'FRAME' || elemTagName == 'IFRAME')
	return 0;

      // for imagemap
      if (elemTagName == 'AREA') {
	var coords = elemArray[hintId].getAttribute("coords").split(",");
	x = Number(coords[0]);
	y = Number(coords[1]);
      }

      //    logMessage("screen:" + window._content.screen.left + ", " + window._content.screen.top);

      //evt.initMouseEvent('click', true, true, window._content, 1, window.screen.left + x+2, window.screen.top + y+2, x+2, y+2, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
      //    evt.layerX = x;
      //    evt.layerY = y;


      // for gmail etc.
      //    if (elem.hasAttribute("onmousedown")) {
      var evt = window._content.document.createEvent('MouseEvents');
      evt.initMouseEvent('mousedown', true, true, doc.defaultView, 1, x+1, y+1, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
      elem.dispatchEvent(evt);
      //    }

      var evt = window._content.document.createEvent('MouseEvents');
      evt.initMouseEvent('click', true, true, doc.defaultView, 1, x+1, y+1, 0, 0, event.ctrlKey, event.altKey, event.shiftKey, event.metaKey, 0, null);
      elem.dispatchEvent(evt);
        
      //doc.documentElement.dispatchEvent(evt);
      //gBrowser.dispatchEvent(evt);

      //    }

      // we don't want to remain the focus for radiobuttons and checkboxes
      if (elemArray[hintId].tagName == 'INPUT') {
	var type = elemArray[hintId].getAttribute("type");
	if (type == "radio" || type == "checkbox")
	  blurFocused();
      }
    } 
    return 0;
  }

  function setMouseOverLink() {
    if (linkNumString.length == 0) return;
  
    var doc = window.document;
    var nums = prefs.getCharPref("keys.numbers");
  
    var hintId = parseInt(linkNumString, nums.length);
  
    //  elem = linksArray[hintId];
    elem = elemArray[hintId];

    //  logMessage(elem.offsetParent.innerHTML)
  
    if (elem.tagName == 'FRAME' || elem.tagName == 'IFRAME') {
      elem.contentWindow.focus();
      return;
    } else {
      elem.focus();
    }
  
    if (hintId >= 0 && hintId < hintElemArray.length) {
      var evt = doc.createEvent('MouseEvents');
      var x = 0;
      var y = 0;
      // for imagemap
      if (elem.tagName == 'AREA') {
	var coords = elem.getAttribute("coords").split(",");
	x = Number(coords[0]);
	y = Number(coords[1]);
      }

      evt.initMouseEvent('mouseover', true, true, doc.defaultView, 1, x, y, 0, 0, 0, 0, 0, 0, 0, null);
      elem.dispatchEvent(evt);
    }
  }

  function charcodeToNumber(charCode) {
    var nums = prefs.getCharPref("keys.numbers").toLowerCase();
    var chr = String.fromCharCode(charCode).toLowerCase();
    var i = nums.indexOf(chr);
    if (i == -1) return null;
    return i.toString(nums.length);
  }

  function updateChanges(linkNumString, oldLinkNumString) {
    changeHintFocus(linkNumString, oldLinkNumString);
    setMouseOverLink();
  }

  function typeNumber(number) {
    if (!isHahModeEnabled) return;

    var oldLinkNumString = null; 
    var nums = prefs.getCharPref("keys.numbers");
  
    if (number != null) {
      oldLinkNumString = linkNumString;
      linkNumString += '' + number;
      var hintId = parseInt(linkNumString, nums.length);
      if (!(hintId >= 0 && hintId < hintElemArray.length))
	linkNumString = oldLinkNumString;
    }
 
    updateChanges(linkNumString, oldLinkNumString);
  }

  function typeBackspace() {
    var oldLinkNumString = null;
    if (linkNumString.length > 0) {
      oldLinkNumString = linkNumString;
      linkNumString = linkNumString.substr(0, linkNumString.length - 1);
    }
    updateChanges(linkNumString, oldLinkNumString);
  }

  ////////////////////////////////////////////////////////////////////////////////
  // event handlers
  ////////////////////////////////////////////////////////////////////////////////

  this.falsifyMagicDoublePress = function() {
    magicDoublePress = false;
  };

  this.enterStartKey = function(event) {
    event.preventDefault();
    event.stopPropagation();
    
    if (!isHahModeEnabled) {
      enableHahMode(event);
    }
    else {
      disableHahMode(event, false);
    }
  };
  
  this.enterMagicKey = function(event) {
    event.preventDefault();
    event.stopPropagation();
    
    if (!isHahModeEnabled) {
      enableHahMode(event);
    }
  };
  
  this.enterBackSpace = function(event) {
    event.preventDefault();
    event.stopPropagation();

    typeBackspace();    
    if (linkNumString == 0) {
      disableHahMode(event, false);
    }
  };
  
  this.releaseMagicKey = function(event) {
    event.preventDefault();
    event.stopPropagation();

    if (isHahModeEnabled) {
      disableHahMode(event, true);
    }

    isMagicKeyPressed = false;
  };

  this.enterExitKey = function(event) {
    event.preventDefault();
    event.stopPropagation();
    
    if (!isHahModeEnabled) {
      blurFocused();
    } else {
      disableHahMode(event, false);
    }
  };

  function onKeydown(event) {
    return;

    var magicDelay = prefs.getIntPref("keys.magicDelay");

    if (magicDoublePress)
      blurFocused();

    magicDoublePress = true;

    window.setTimeout("hah.falsifyMagicDoublePress();", magicDelay);
  }
  
  function onKeyup(event) {
    
    var magicKey = window.document.getElementById("hahMagicKey").getAttribute("key");
    
    var eventKey;

    if (magicKey.length == 1) {
      eventKey = String.fromCharCode(event.keyCode);
    } else {
      eventKey = keycodeToName(event.keyCode);
    }

    if (!isHahModeEnabled) return;
    
    if (eventKey == magicKey) {
      hah.releaseMagicKey(event);
    }
  }

  function keycodeToName(keyCode) {
    for (keyName in KeyboardEvent.prototype) {
      if (keyName.substr(0,7) == 'DOM_VK_' && KeyboardEvent.prototype[keyName] == keyCode) {
	// turn 'DOM_VK_KEY_NAME' into 'Key Name'
	parts = keyName.substr(7).split('_'); // strip off DOM_VK_ and split into words
	for (var i = 0; i < parts.length; i++)
	  parts[i] = parts[i][0] + parts[i].substr(1).toLowerCase();
	return parts.join(' ');
      }
    }
    return 0
  }

  function onKeypress(event) {
    // if the frame has been zoomed in/out
    if ((event.charCode == 45 || event.charCode == 43) && event.ctrlKey) {
      var p = false;
      disableHahMode();
      resize();
    }

    if (!isHahModeEnabled) return;

    var num = charcodeToNumber(event.charCode);

    if (num != null) {
      event.preventDefault();
      event.stopPropagation(); 
      typeNumber(num);
    }
  }

  window.document.addEventListener("DOMContentLoaded", load, null);

  //window.addEventListener("load", load, true);
  window.document.addEventListener("resize", resize, true);

  //is there any method to do this?
  //would be nice with js popup menus etc.
  //window.document.addEventListener("mutation", hahMutate, true);

  // temporal solution
  //window.document.addEventListener("click", resize, true);
  window.document.addEventListener("change", resize, true);

  //window.document.addEventListener("zoom", resize, true);
  //window.addEventListener("blur", disableHahMode, true);
  //window.onblur = disableHahMode;

  logMessage("Initialized.");

}

var hah = new hit_a_hint();
